/*

  Dumbug tracer 

  */
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <conio.h>
#include <time.h>
#include <Tlhelp32.h>

#include <map>
#include <vector>
using namespace std;

// project elements
#include "debugger.hpp"
#include "log.hpp"
#include "pecoff.hpp"
#include "tracer.hpp"
#include "parser/tracedefs.hpp"


#define DEMO_VERSION "v0.3 (tracer/fork)"


//
// Global, so the handlers can access it
//
bool		MainTerminate = false;
Tracer		dbg;

//
// for nowin's child follow
//
vector <Debugger*>		Children;

//
// callback functions
//
void itsdone(DEBUG_EVENT *db);
void initial_break(DEBUG_EVENT *db);
void normal_break(DEBUG_EVENT *db);
void openr(t_Tracer_OpenTr *open);
void closer(t_Tracer_OpenTr *open);

//
// local prototypes
//
void dump_results(FILE *fd);
void dump_arg(FILE *fd, t_Tracedef_Argument *a);
void list_processes();


int main(int argc, char **argv) {
	FILE		*resfile;
	int			cit;

	//
	// Set the debugger's log level
	//
	dbg.log.set(LOG_INFO);
	dbg.SetGlobalLoglevel(LOG_INFO);

	if (2 > argc) {
		printf("Usage: %s <[PID]|filename.exe> [<trace.def>] [<traceout.txt>]\n",argv[0]);
		list_processes();
		return 1;
	} else {
		//
		// check if numeric, then attach
		//
		if ( 0 == atoi(argv[1]) ) {
			//
			// try to load file
			//
			if (!dbg.load(argv[1])) {
				// 
				// didn't work, try attaching to named process child
				//
				if ( !dbg.attach(argv[1],true) ) {
					//
					// didn't work either, so look for a lonely parrent
					//
					if ( !dbg.attach(argv[1],false) ) {
						printf("Could neither load nor attach to something called %s\n",
							argv[1]);
						return 1;
					}
				}
			}
		} else {
			//
			// numeric argument, attach
			//
			if (!dbg.attach(atoi(argv[1]))) {
				printf("No process with the ID %u\n", atoi(argv[1]));
				return 1;
			}
		}

		//
		// More arguments mean trace definition
		//
		if ( 3 <= argc ) {
			dbg.LoadTraceDefinitions(argv[2]);
		}

		//
		// Even more arguments mean trace dump file
		//
		if ( 4 <= argc ) {
			if ( NULL == (resfile=fopen(argv[3],"wt")) ){
				printf("Error opening file %s\n",argv[3]);
				return 2;
			}
		} else {
			resfile = stdout;
		}
	}

	//
	// assign functions to handle breakpoints
	//
	dbg.set_inital_breakpoint_handler(&initial_break);
	//dbg.set_breakpoint_handler(&normal_break);
	dbg.SetOpenRecordFunction(&openr);
	dbg.SetCloseRecordFunction(&closer);

	//
	// assign a function to handle exit event
	//
	dbg.set_exit_handler(&itsdone);

	printf(
		"=============================================================\n"
		"(dum(b)ug) demo %s\n"
		"[q] Quit  [I] Imports  [E] Exports  [M] MemoryMap  [T] Traces\n"
		"[1]-[0] Log Level (1-5 Debugger, 6-9 Tracer, 1/6 = LOG_DEBUG)\n"
		"=============================================================\n",
		DEMO_VERSION
		);

	dbg.Analysis.StackDelta = true;
	dbg.Analysis.StackDeltaV = 2048;

	// 
	// application main loop
	//
	while (!MainTerminate) {
		
		dbg.run(1000);

		for (cit=0; cit < Children.size(); cit++) {
			Children[cit]->run(1000);
		}

		if (_kbhit()) {
			switch(getch()) {
				case 'q':	
					MainTerminate=true;
					break;
				case 'E': 
					{
						int	i,j;
						
						printf("Export tables:\n");
						
						for (i=0; i<dbg.Files.size(); i++) {
							for (j=0; j<dbg.Files[i]->exports.size(); j++) {

								if (dbg.Files[i]->exports[j]->Forwarded) {
									printf("  %s: %08X %s (%u) -> %s\n",
										dbg.Files[i]->internal_name.c_str(),
										0,
										dbg.Files[i]->exports[j]->Name.c_str(),
										dbg.Files[i]->exports[j]->Ordinal,
										dbg.Files[i]->exports[j]->ForwardName.c_str());
								} else {
									printf("  %s: %08X %s (%u)\n", 
										dbg.Files[i]->internal_name.c_str(),
										dbg.Files[i]->exports[j]->Address,
										dbg.Files[i]->exports[j]->Name.c_str(),
										dbg.Files[i]->exports[j]->Ordinal);
								}
							}
						}
					}
					break;
				case 'I':
					{
						int i,j,k;

						printf("Import tables:\n");

						printf("%s: \n",dbg.getImageName());
						
						// Import table of the PEimage itself
						for (j=0; j<dbg.PEimage.imports.size(); j++) {
							
							printf("  %s\n",dbg.PEimage.imports[j]->Name.c_str());
							
							for (k=0; k<dbg.PEimage.imports[j]->functions.size(); k++) {
								printf("    %08X %s\n",
									dbg.PEimage.imports[j]->functions[k]->Address,
									dbg.PEimage.imports[j]->functions[k]->Name.c_str());
							}
						}

						// Import table of the DLLs
						for (i=0; i<dbg.Files.size(); i++) {

							printf("%s: \n",dbg.Files[i]->internal_name.c_str());

							for (j=0; j<dbg.Files[i]->imports.size(); j++) {

								printf("  %s\n",dbg.Files[i]->imports[j]->Name.c_str());

								for (k=0; k<dbg.Files[i]->imports[j]->functions.size(); k++) {
									printf("    %08X %s\n",
										dbg.Files[i]->imports[j]->functions[k]->Address,
										dbg.Files[i]->imports[j]->functions[k]->Name.c_str());
								}
							}
						}
					}
					break;

				case 'M':
					{
						map <DWORD,t_Debugger_memory*>::const_iterator	it;

						printf("Memory Map:\n");
						for ( it = dbg.MemoryMap.begin();
							it != dbg.MemoryMap.end();
							++it ){
							printf("  %08X %10u %s (protection %08X)\n",
								it->first,
								it->second->basics.RegionSize,
								it->second->Name.c_str(),
								it->second->basics.Protect);
						}
					}
					break;

				case 'T': 
					{
						t_Tracevector_Iterator		ot;
						string						trstr;

						printf("Trace definitions:\n");
						for ( ot = dbg.Traces.begin(); 
							ot < dbg.Traces.end();
							ot ++) {
								dbg.GetTraceString( &(*ot), &trstr);
								printf("  %s\n",trstr.c_str());
							}
					}
					break;

				case 'r':
					dump_results(stdout);
					break;

				//
				// Changing Log levels
				//
				case '1':
					dbg.SetGlobalLoglevel(LOG_DEBUG);
					break;
				case '2':
					dbg.SetGlobalLoglevel(LOG_VERBOSE);
					break;
				case '3':
					dbg.SetGlobalLoglevel(LOG_INFO);
					break;
				case '4':
					dbg.SetGlobalLoglevel(LOG_WARNING);
					break;
				case '5':
					dbg.SetGlobalLoglevel(LOG_ERROR);
					break;
				case '6':
					dbg.log.set(LOG_DEBUG);
					break;
				case '7':
					dbg.log.set(LOG_VERBOSE);
					break;
				case '8':
					dbg.log.set(LOG_INFO);
					break;
				case '9':
					dbg.log.set(LOG_WARNING);
					break;
				case '0':
					dbg.log.set(LOG_ERROR);
					break;

			} // switch
		}
	}

	dump_results(resfile);

	//
	// no cleanup required, Debugger class will handle it
	//
	
	return 0;
};



void initial_break(DEBUG_EVENT *db) {

	printf("[--initial--] Breakpoint - activating traces\n");
	dbg.ActivateTraces();

}


void normal_break(DEBUG_EVENT *db) {
	t_disassembly		dis;


	printf("[--normal--] Breakpoint\n");
	
	if ( dbg.Disassemble(
		dbg.FindThread( (DWORD)db->dwThreadId )->hThread,
		(DWORD)(db->u.Exception.ExceptionRecord.ExceptionAddress),
		&dis)) {

		printf("%08X   %s\n",
			(DWORD)(db->u.Exception.ExceptionRecord.ExceptionAddress),
			dis.mnemonic.c_str());
	}
}


void itsdone(DEBUG_EVENT *db) {
	printf("Debuggee terminated, doing the same\n");
	MainTerminate = true;
}


void openr(t_Tracer_OpenTr *open) {
	// do mothing
}


void closer(t_Tracer_OpenTr *open) {
	Debugger				*c;
	PROCESS_INFORMATION		ProcInfo;
	char					bla[30];
	
	if ( open->Definition->Name == "CreateProcessA" ) {
		printf("MAIN: Process created (CreateProcessA)\n");

		if ( ! ReadProcessMemory(dbg.getProcessHandle(),
			(LPCVOID)(open->OutArgs[9].data.address),
			&ProcInfo,
			sizeof(ProcInfo),
			NULL) ) {
			printf("Failed to read process memory at %08X\n",
				open->OutArgs[9].data.address);
		} else {
			c = new (Debugger);

			sprintf(bla,"Child %u",ProcInfo.dwProcessId);
			c->attach(ProcInfo.dwProcessId);
			c->log.Name= bla;

			Children.push_back(c);
		}
	} else if ( open->Definition->Name == "CreateProcessAsUserA" ) {
		printf("MAIN: Process created (CreateProcessAsUserA)\n");

		if ( ! ReadProcessMemory(dbg.getProcessHandle(),
			(LPCVOID)(open->OutArgs[10].data.address),
			&ProcInfo,
			sizeof(ProcInfo),
			NULL) ) {
			printf("Failed to read process memory at %08X\n",
				open->OutArgs[10].data.address);
		} else {
			c = new (Debugger);

			sprintf(bla,"Child %u",ProcInfo.dwProcessId);
			c->attach(ProcInfo.dwProcessId);
			c->log.Name= bla;

			Children.push_back(c);
		}
	} 
}


void dump_results(FILE *fd) {
	time_t						t;
	t_Resultvector_Iterator		r;
	t_Argumentvector_Iterator	a;


	printf("Dumping %u results...\n",dbg.Results.size());

	t=time(NULL);
	fprintf(fd,"\n\n%sTrace results:\n\n\n",ctime(&t));

	r=dbg.Results.begin();
	while ( r != dbg.Results.end() ) {

		fprintf(fd,"------------------------------------------------------\n");
		fprintf(fd,"%08X -> %s(\n", (*r)->caller, (*r)->Definition->Name.c_str() );
		// before

		a = (*r)->InArgs.begin();
		while ( a != (*r)->InArgs.end() ) {
			dump_arg( fd, &(*a) );

			a++;
			
			if ( a != (*r)->OutArgs.end() )
				fprintf(fd,",\n");
			else
				fprintf(fd,"\n);\n");
		}

		//
		// after
		//
		switch ( (*r)->Definition->Return ) {
		case tr_VOID:
			fprintf(fd,"%8s",""); 
			break;
		case tr_INT:
			fprintf(fd,"%8d", (*r)->Return.value );
			break;
		case tr_CHAR:
			fprintf(fd,"0x%02X    ", (*r)->Return.c );
			break;
		case tr_PVOID:
		case tr_PINT:
		case tr_PCHAR:
		case tr_PFMTCHAR:
		case tr_PWCHAR:
		case tr_PFMTWCHAR:
			fprintf(fd,"%08X", (*r)->Return.address );
			break;
		default:
			fprintf(fd,"????????");
		}

		fprintf(fd," << %s(\n",(*r)->Definition->Name.c_str() );

		a = (*r)->OutArgs.begin();
		while ( a != (*r)->OutArgs.end() ) {
			dump_arg( fd, &(*a) );

			a++;
			
			if ( a != (*r)->OutArgs.end() )
				fprintf(fd,",\n");
			else
				fprintf(fd,"\n);\n");
		}

		if ( (*r)->Notes.length() > 0 )
			fprintf(fd,"\nNotes:\n%s",(*r)->Notes.c_str());

		// next result
		r++;
	} // while results
}


void dump_arg(FILE *fd, t_Tracedef_Argument *a) {

	fprintf(fd,"\t%-9s %10s =", GetTypeName( a->Type ), a->Name.c_str());
	switch ( a->Type ) {
	case tr_VOID:	
		fprintf(fd," (big black nothing) ");
		break;
	case tr_INT:	
		fprintf(fd," %d (unsigned = %u / hex = %X)", a->data.value,
			a->data.value,a->data.value);
		break;
	case tr_CHAR:
		fprintf(fd," %02X ('%c')",a->data.c, a->data.c);
		break;
	case tr_PVOID:
	case tr_PINT:
		fprintf(fd," %08X",a->data.address);
		break;
	case tr_PCHAR:
	case tr_PFMTCHAR:
	case tr_PWCHAR:
	case tr_PFMTWCHAR:
		fprintf(fd," [%08X] = %s",a->data.address,a->comment.c_str());
		break;
	default:
		fprintf(fd," eh???");
	} // switch type

}


void list_processes() {
	HANDLE			snap;
	PROCESSENTRY32	proc;
	char			drive[_MAX_DRIVE];
	char			dir[_MAX_DIR];
	char			fname[_MAX_FNAME];
	char			fnameP[_MAX_FNAME];
	char			ext[_MAX_EXT];


	if ( INVALID_HANDLE_VALUE == 
		(snap=CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS,0)) ){
		return;
	}

	proc.dwSize = sizeof(proc);
	if ( ! Process32First(snap,&proc) ) {
		CloseHandle(snap);
		return;
	} 

	printf("\nCurrent Processes, FYI:\n%8s %8s  %s\n","PID","PPID","Name");
	do {
		_splitpath(proc.szExeFile,drive,dir,fname,ext);
		printf("%8u %8u  %s\n",proc.th32ProcessID,proc.th32ParentProcessID,fname);
	} while ( Process32Next(snap,&proc) );

	CloseHandle(snap);
}